// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Congestion Manager Client Implementation                         ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "timedthread.h"
#include "socket.h"
#include "internetaddress.h"
#include "transportinfo.h"
#include "congestionmonitorreport.h"
#include "congestionmanagerpacket.h"
#include "tools.h"
#include "congestionmanagerclient.h"


// Verbose mode: Display some information on UDP connections.
#define VERBOSE


namespace Coral {


// ###### Constructor #######################################################
CongestionManagerClient::CongestionManagerClient(const char* manager)
   : TimedThread(250000)
{
   Ready = false;
   setTimerCorrection(0);
   SequenceNumber = 0;

   // ====== Create connection to manager ===================================
   ManagerSocket.create(Socket::IP,Socket::UDP);
   if(!ManagerSocket.ready()) {
      cerr << "ERROR: CongestionManagerClient::CongestionManagerClient() - "
              "Unable to create socket!" << endl;
      return;
   }

   InternetAddress managerAddress(manager);
   Flow = ManagerSocket.allocFlow(managerAddress);
   if(Flow.getFlowLabel() != 0) {
      if(ManagerSocket.connect(Flow,CongestionManagerClientTrafficClass) == 0) {
         if(ManagerSocket.connect(managerAddress,CongestionManagerClientTrafficClass) == false) {
            cerr << "ERROR: CongestionManagerClient::CongestionManagerClient() - "
                    "Unable to flow-connect to congestion manager!" << endl;
            return;
         }
      }
   }
   else {
      if(ManagerSocket.connect(managerAddress,CongestionManagerClientTrafficClass) == false) {
         cerr << "ERROR: CongestionManagerClient::CongestionManagerClient() - "
                 "Unable to connect to congestion manager!" << endl;
         return;
      }
   }
   ManagerSocket.setBlockingMode(false);

   // ====== Create socket for replies ======================================
   ReplySocket.create(Socket::IP,Socket::UDP);
   if(!ReplySocket.ready()) {
      cerr << "ERROR: CongestionManagerClient::CongestionManagerClient() - "
              "Unable to create socket!" << endl;
      return;
   }
   InternetAddress address;
   ManagerSocket.getSocketAddress(address);
   address.setPort(0);
   if(ReplySocket.bind(address) == false) {
      cerr << "ERROR: CongestionManagerClient::CongestionManagerClient() - Unable to bind socket for replies!" << endl;
      return;
   }
   ReplySocket.setBlockingMode(false);
   ReplySocket.getSocketAddress(address);
   ReplyAddress = address.getPortableAddress();

   // ====== Print information ==============================================
#ifdef VERBOSE
   ManagerSocket.getPeerAddress(address);
   cout << "Connection to congestion manager " << address;
   if(ManagerSocket.getSendFlowLabel()) {
      char str[128];
      snprintf((char*)&str,sizeof(str),"$%05x, traffic class $%02x.",
              ManagerSocket.getSendFlowLabel(),ManagerSocket.getSendTrafficClass());
      cout << " IPv6 flow label is " << str << endl;
   }  
   else {
      cout << "." << endl;
   }
#endif
   Ready = true;
}


// ###### Destructor ########################################################
CongestionManagerClient::~CongestionManagerClient()
{
   stop();

   if(Flow.getFlowLabel() != 0) {
      ManagerSocket.freeFlow(Flow);
   }

   multimap<const cardinal,ManagedStream*>::iterator streamIterator = 
      StreamSet.begin();
   while(streamIterator != StreamSet.end()) {
      ManagedStream* stream = streamIterator->second;
      StreamSet.erase(streamIterator);
      delete stream;
      streamIterator = StreamSet.begin();
   }
}


// ###### Add stream to management ##########################################
bool CongestionManagerClient::addStream(const cardinal               identifier,
                                        const ExtendedTransportInfo& streamDescription)
{
   // ====== Check, if client is ready ======================================
   if(!ready()) {
      cerr << "ERROR: CongestionManagerClient::addStream() - CongestionManagerClient not ready!"
           << endl;
      return(false);
   }

   // ====== Create new member ==============================================
   synchronized();
   multimap<const cardinal,ManagedStream*>::iterator found = StreamSet.find(identifier);
   if(found != StreamSet.end()) {
      unsynchronized();
      cerr << "WARNING: CongestionManagerClient::addStream() - Identifier " << identifier
           << " already used!" << endl;
      return(false);
   }

   ManagedStream* stream = new ManagedStream;
   if(stream == NULL) {
      unsynchronized();
      cerr << "ERROR: CongestionManagerClient::addStream() - Out of memory!" << endl;
      return(false);
   }
   stream->StreamDescription = streamDescription;
   stream->SequenceNumber    = 0;
   stream->LastSuggestion    = 0;

   // New member will start with minimum bandwidth until we receive a report
   // from manager!
   stream->Suggestion        = streamDescription;
   stream->Suggestion.WantedBytesPerSecond[0].setValue(stream->Suggestion.WantedBytesPerSecond[0].getMin());
   StreamSet.insert(pair<const cardinal,ManagedStream*>(identifier,stream));

   unsynchronized();
   return(true);
}


// ###### Remove stream from management #####################################
void CongestionManagerClient::removeStream(const cardinal identifier)
{
   synchronized();
   multimap<const cardinal,ManagedStream*>::iterator found = StreamSet.find(identifier);
   if(found != StreamSet.end()) {
      ManagedStream* stream = found->second;

      // ====== Send Bye ====================================================
      CongestionManagerRequest request;
      request.PacketID       = CongestionManagerPacket::CongestionManagerPacketID;
      request.SequenceNumber = stream->SequenceNumber;
      request.Type           = CongestionManagerPacket::CMPT_Bye;
      request.Length         = sizeof(CongestionManagerRequest);
      InternetAddress address;
      ReplySocket.getSocketAddress(address);
      request.ReplyTo        = address.getPortableAddress();
      request.Identifier        = identifier;
      request.StreamDescription = stream->StreamDescription;
      request.translate();
      ManagerSocket.send(&request,sizeof(request));

      // ====== Remove stream ===============================================
      StreamSet.erase(found);
      delete stream;
   }
   else {
      cerr << "WARNING: CongestionManagerClient::removeStream() - Invalid identifier!"
           << endl;
   }
   unsynchronized();
}


// ###### Get suggested transport info ######################################
bool CongestionManagerClient::getSuggestedTransportInfo(const cardinal         identifier,
                                                        ExtendedTransportInfo& streamDescription)
{
   bool result = false;
   synchronized();
   multimap<const cardinal,ManagedStream*>::iterator found = StreamSet.find(identifier);
   if(found != StreamSet.end()) {
      ManagedStream* stream = found->second;
      streamDescription = stream->Suggestion;
      result = true;
   }
   else {
      cerr << "WARNING: CongestionManagerClient::getSuggestedTransportInfo() - Invalid identifier!"
           << endl;
   }
   unsynchronized();
   return(result);
}


// ###### Set transport info ################################################
void CongestionManagerClient::setTransportInfo(const cardinal               identifier,
                                               const ExtendedTransportInfo& streamDescription)
{
   synchronized();

   multimap<const cardinal,ManagedStream*>::iterator found = StreamSet.find(identifier);
   if(found != StreamSet.end()) {
      ManagedStream* stream = found->second;
      stream->StreamDescription = streamDescription;
      stream->SequenceNumber++;
   }
   else {
      cerr << "WARNING: CongestionManagerClient::setTransportInfo() - Invalid identifier!"
           << endl;
   }

   unsynchronized();
}


// ###### Send requests #####################################################
void CongestionManagerClient::sendRequests()
{   
   multimap<const cardinal,ManagedStream*>::iterator streamIterator = 
      StreamSet.begin();

   while(streamIterator != StreamSet.end()) {
      const cardinal identifier = streamIterator->first;
      ManagedStream* stream     = streamIterator->second;

      // ====== Send Request ================================================
      CongestionManagerRequest request;
      request.PacketID          = CongestionManagerPacket::CongestionManagerPacketID;
      request.SequenceNumber    = stream->SequenceNumber;
      request.Type              = CongestionManagerPacket::CMPT_Request;
      request.Length            = sizeof(CongestionManagerRequest);
      request.ReplyTo           = ReplyAddress;
      request.Identifier        = identifier;
      request.StreamDescription = stream->StreamDescription;
      request.translate();
      ManagerSocket.send(&request,sizeof(request));

      // ====== Get next stream =============================================
      streamIterator++;
   }
}


// ###### Receive responses #################################################
void CongestionManagerClient::receiveResponses()
{
   CongestionManagerPacket packet;

   // ====== Get packet header from socket ==================================
   int32 packetSize = ReplySocket.receive(&packet,
                         sizeof(CongestionManagerPacket),MSG_PEEK);
   while(packetSize >= (int32)sizeof(CongestionManagerPacket)) {
      char   packetData[packet.Length];
      cardinal receivedPacketSize = ReplySocket.receive(
                                       (char*)&packetData,packet.Length);
      if(translate16(packet.PacketID) == CongestionManagerPacket::CongestionManagerPacketID) {

         // ====== Packet is a CMPT_Request =================================
         switch(translate16(packet.Type)) {
            case CongestionManagerPacket::CMPT_Suggestion:
               if(receivedPacketSize >= sizeof(CongestionManagerSuggestion)) {
                  CongestionManagerSuggestion* suggestion = (CongestionManagerSuggestion*)packetData;
                  suggestion->translate();
                  handleSuggestion(suggestion);
               }
             break;
            default:
                cerr << "Unknown packet type " << translate16(packet.Type) <<endl;
             break;
         }
      }
      
      // ====== Get next packet header from socket ==========================
      packetSize = ReplySocket.receive(&packet,sizeof(CongestionManagerPacket),MSG_PEEK);
   }
}


// ###### Handle bandwidth suggestion #######################################
void CongestionManagerClient::handleSuggestion(const CongestionManagerSuggestion* suggestion)
{
   synchronized();
   multimap<const cardinal,ManagedStream*>::iterator found =
      StreamSet.find(suggestion->Identifier);
   if(found != StreamSet.end()) {
      ManagedStream* stream = found->second;

      if(suggestion->SequenceNumber >= stream->LastSuggestion) {
         // ?????? SeqNumValidator hier !!!!!

         stream->Suggestion     = suggestion->Suggestion;
         stream->LastSuggestion = suggestion->SequenceNumber;
      }
      else {
         // cout << "Outdated suggestion ignored." << endl;
      }
   }
   unsynchronized();
}


// ###### TimedThread's timerEvent() implementation #########################
void CongestionManagerClient::timerEvent()
{
   // Send requests and get suggestions
   if(Ready) {
      synchronized();
      sendRequests();
      receiveResponses();
      unsynchronized();
   }
}


}
