// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Congestion Monitor Simulator Implementation                      ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "socket.h"
#include "internetaddress.h"
#include "congestionmonitor.h"
#include "congestionmonitorreport.h"
#include "networkmonitorsimulator.h"
#include "tools.h"
#include "csim.h"


#include <qapp.h>
#if QT_VERSION < 200
#error QT Version 2.00 or better required for QCongestionMonitorSimulator!!!
#endif
#include <qlayout.h>
#include <qpushbutton.h>
#include <qscrollbar.h>
#include <qlabel.h>
#include <qlcdnumber.h>
#include <qgroupbox.h>
#include <qplatinumstyle.h>
#include <qmainwindow.h>


#include <signal.h>


// ====== Include Qt's meta objects ======
#include "csim.moc"


// ###### Constructor #######################################################
QCongestionMonitorSimulator::QCongestionMonitorSimulator(NetworkMonitorSimulator* simulator,
                                           const card64             bandwidth,
                                           QWidget*                 parent,
                                           const char*              name)
   : QMainWindow(parent,name)
{
   Bandwidth = bandwidth;
   Simulator = simulator;

   // ====== Initialize status display ======================================
   QWidget* centralWidget = new QWidget(this);
   QVBoxLayout* layout    = new QVBoxLayout(centralWidget);
   
   QGroupBox* statusGroup    = new QGroupBox("Simulated Congestion",centralWidget);
   QGridLayout* statusLayout = new QGridLayout(
                                      statusGroup,
                                      sizeof(QCongestionMonitorSimulatorStatusLabels) / sizeof(char*),
                                      2,20);
   for(cardinal i = 0;i < sizeof(QCongestionMonitorSimulatorStatusLabels) / sizeof(char*);i++) {
      QLabel* label = new QLabel(statusGroup);
      label->setText(QCongestionMonitorSimulatorStatusLabels[i]);
      label->setMinimumSize(label->sizeHint());
      statusLayout->addWidget(label,i,0);

      Labels[i] = new QLabel("0",statusGroup);
      Labels[i]->setMinimumSize(Labels[i]->sizeHint());
      statusLayout->addWidget(Labels[i],i,1);
   }
   char str[64];
   snprintf((char*)&str,sizeof(str),"%Ld bit/s",bandwidth * 8);
   Labels[2]->setText((char*)&str);
   layout->addWidget(statusGroup);

   // ====== Add scrollbar and button =======================================
   QScrollBar* scrollBar = new QScrollBar(QScrollBar::Horizontal,centralWidget);
   scrollBar->setRange(0,1000);
   QPushButton* button   = new QPushButton("Quit",centralWidget);
   QLabel* copyright     = new QLabel("Copyright (C) 1999-2001 Thomas Dreibholz",centralWidget);

   layout->addWidget(scrollBar);
   layout->addWidget(button);
   layout->addWidget(copyright);

   // ====== Connect Qt slots ===============================================
   QObject::connect(button,SIGNAL(clicked()),this,SLOT(quit()));
   QObject::connect(scrollBar,SIGNAL(valueChanged(int)),this,SLOT(congestion(int)));   
   setCentralWidget(centralWidget);
   congestion(0);
}


// ###### Destructor ########################################################
QCongestionMonitorSimulator::~QCongestionMonitorSimulator()
{
}


// ###### Quit ##############################################################
void QCongestionMonitorSimulator::quit()
{
   exit(0);
}


// ###### Set new congestion value ##########################################
void QCongestionMonitorSimulator::congestion(int value)
{
   Simulator->setCongestion(value / 10.0);
   char str[64];
   snprintf((char*)&str,sizeof(str),"%1.1f %%",value / 10.0);
   Labels[0]->setText((char*)&str);
   snprintf((char*)&str,sizeof(str),"%Ld bit/s",(card64)(Bandwidth * 8.0 * (value / 1000.0)));
   Labels[1]->setText((char*)&str);
}



/* Thread for accepting TCP connections */
class AcceptReceiver : Thread
{
   // ====== Constructor ====================================================
   public:
   AcceptReceiver(CongestionMonitor* monitor, Socket* socket);

   // ====== Private data ===================================================
   private:
   void run();

   CongestionMonitor* Monitor;
   Socket*            AcceptSocket;
};


// ###### Constructor #######################################################
AcceptReceiver::AcceptReceiver(CongestionMonitor* monitor, Socket* socket)
   : Thread()
{
   Monitor = monitor;
   AcceptSocket = socket;
   start();
}


// ###### Thread's run() implementation #####################################
void AcceptReceiver::run()
{
   AcceptSocket->listen(5);
   for(;;) {
      Socket* s = AcceptSocket->accept();
      if(s != NULL) {
         Monitor->addReportReceiver(s);
      }
   }
}



// ###### Main program ######################################################
int main(int argc, char** argv)
{
   char*  local     = NULL;
   card64 interval  = 1000000;
   card64 bandwidth = 0;
   card16 port      = 0;

   // ====== Get arguments ==================================================
   if(argc < 2) {
      cerr << "Usage: " << argv[0] << endl
           << "       [Port] {-bandwidth=bit/s} {-interval=ms} {-local=hostname}"
           << endl;
      exit(1);
   }
   for(cardinal i = 2;i < (cardinal)argc;i++) {
      if(!(strncasecmp(argv[i],"-bandwidth=",11))) {
          bandwidth = atol(&argv[i][11]);
      }
      else if(!(strncasecmp(argv[i],"-interval=",10))) {
         interval = (card64)atol(&argv[i][10]) * 1000;
      }
      else if(!(strncasecmp(argv[i],"-local=",7))) {
         local = &argv[i][7];
      }
      else {
         cerr << "Wrong parameter: " << argv[i] << endl;
         cerr << "Usage: " << argv[0] << endl
              << "       [Port] {-bandwidth=bit/s} {-interval=ms} {-local=hostname}"
              << endl;
         exit(1);
      } 
   }
   port = atoi(argv[1]);
   if(bandwidth < 1000) {
      cerr << "WARNING: No bandwidth given. Using default 10 MBit/s!" << endl;
      bandwidth = 10000000;
   }
   if(interval < 1000)
      interval = 1000;
   if(interval > 60000000)
      interval = 60000000;


   // ====== Create socket for report receivers =============================
   InternetAddress monitorAddress(local,port);
   Socket          monitorSocket(Socket::IP,Socket::TCP);
   monitorSocket.setSoReuseAddress(true);
   monitorSocket.setSoLinger(true,3);
   if(monitorSocket.bind(monitorAddress) == false) {
      cerr << "ERROR: Bind to " << monitorAddress << " failed!" << endl;
      exit(1);
   }
   monitorSocket.listen(5);

   // ====== Initialize Qt ==================================================
   QApplication* application = new QApplication(argc,argv);
   application->setStyle(new QPlatinumStyle());

   // ====== Start congestion monitor =======================================
   NetworkMonitorSimulator* netMonitor = new NetworkMonitorSimulator(bandwidth / 8);
   if(netMonitor->ready()) {
      // SIGPIPE has to be ignored!! It is sent, if a TCP connection is
      // closed and a write() is tried -> Program terminates with SIGPIPE.
      // But we want the monitor only to remove the receiver. Therefore:
      signal(SIGPIPE,SIG_IGN);

      CongestionMonitor* cgMonitor = new CongestionMonitor(netMonitor,bandwidth / 8,CMF_SIMULATION);
      if(cgMonitor != NULL) {
         AcceptReceiver acceptor(cgMonitor,&monitorSocket);
         cgMonitor->setInterval(interval);
         cgMonitor->start();

         // ====== Startin GUI ==============================================
         QCongestionMonitorSimulator* simulator = new QCongestionMonitorSimulator(
                                                      netMonitor,bandwidth / 8);
         if(simulator != NULL) {
            application->setMainWidget(simulator);
            simulator->show();
            application->exec();
         }

         // ====== Clean-up ====================================================
         cgMonitor->stop();
         delete cgMonitor;
      }
      delete netMonitor;
   }
   return(0);
}
