// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Multi Audio Writer                                               ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include <iostream.h>
#include <sys/time.h>


#include "system.h"
#include "multiaudiowriter.h"
#include "tools.h"


namespace Coral {


// ###### Constructor #######################################################
MultiAudioWriter::MultiAudioWriter()
   : Synchronizable()
{
   AudioSamplingRate = 44100;
   AudioBits         = 16;
   AudioChannels     = 2;
}


// ###### Destructor ########################################################
MultiAudioWriter::~MultiAudioWriter()
{
}


// ###### Add writer to writer set ##########################################
bool MultiAudioWriter::addWriter(AudioWriterInterface* writer)
{
   synchronized();
   WriterSet.insert(writer);
   writer->setQuality(*this);
   unsynchronized();
   return(true);
}


// ###### Remove writer from writer set #####################################
void MultiAudioWriter::removeWriter(AudioWriterInterface* writer)
{
   synchronized();
   multiset<AudioWriterInterface*>::iterator found = WriterSet.find(writer);
   if(found != WriterSet.end()) {
      WriterSet.erase(found);
   }
   unsynchronized();
}


// ###### Get number of channels ############################################
card8 MultiAudioWriter::getChannels() const
{
   return(AudioChannels);
}


// ###### Get number of bits ################################################
card8 MultiAudioWriter::getBits() const
{
   return(AudioBits);
}


// ###### Get sampling rate #################################################
card16 MultiAudioWriter::getSamplingRate() const
{
   return(AudioSamplingRate);
}


// ###### Get byte order ####################################################
card16 MultiAudioWriter::getByteOrder() const
{
   return(AudioByteOrder);
}


// ###### Set number of bits ################################################
card8 MultiAudioWriter::setBits(const card8 bits) {
   synchronized();
   AudioBits = bits;

   multiset<AudioWriterInterface*>::iterator writerIterator = WriterSet.begin();
   while(writerIterator != WriterSet.end()) {
      (*writerIterator)->setBits(AudioBits);
      writerIterator++;
   }

   unsynchronized();
   return(AudioBits);
}


// ###### Set number of channels ############################################
card8 MultiAudioWriter::setChannels(const card8 channels) {
   synchronized();
   AudioChannels = channels;

   multiset<AudioWriterInterface*>::iterator writerIterator = WriterSet.begin();
   while(writerIterator != WriterSet.end()) {
      (*writerIterator)->setChannels(AudioChannels);
      writerIterator++;
   }

   unsynchronized();
   return(AudioChannels);
}


// ###### Set sampling rate #################################################
card16 MultiAudioWriter::setSamplingRate(const card16 rate) {
   synchronized();
   AudioSamplingRate = rate;  

   multiset<AudioWriterInterface*>::iterator writerIterator = WriterSet.begin();
   while(writerIterator != WriterSet.end()) {
      (*writerIterator)->setSamplingRate(AudioSamplingRate);
      writerIterator++;
   }

   unsynchronized();
   return(AudioSamplingRate);
}


// ###### Set byte order ####################################################
card16 MultiAudioWriter::setByteOrder(const card16 byteOrder)
{
   synchronized();
   AudioByteOrder = byteOrder;

   multiset<AudioWriterInterface*>::iterator writerIterator = WriterSet.begin();
   while(writerIterator != WriterSet.end()) {
      (*writerIterator)->setByteOrder(AudioByteOrder);
      writerIterator++;
   }

   unsynchronized();
   return(AudioByteOrder);
}


// ###### Get bytes per second ##############################################
cardinal MultiAudioWriter::getBytesPerSecond() const
{
   return((AudioSamplingRate * AudioChannels * AudioBits) / 8);
}


// ###### Get bits per sample ###############################################
cardinal MultiAudioWriter::getBitsPerSample() const
{
   return(AudioChannels * AudioBits);
}


// ###### Check, if spectrum analyzer is ready ##############################
bool MultiAudioWriter::ready() const
{
   return(true);
}


// ###### Reset #############################################################
void MultiAudioWriter::sync()
{
   synchronized();

   multiset<AudioWriterInterface*>::iterator writerIterator = WriterSet.begin();
   while(writerIterator != WriterSet.end()) {
      (*writerIterator)->sync();
      writerIterator++;
   }

   unsynchronized();
}


// ###### Write data to spectrum analyzer ###################################
bool MultiAudioWriter::write(const void* data, const size_t length)
{
   bool success = true;
   synchronized();

   multiset<AudioWriterInterface*>::iterator writerIterator = WriterSet.begin();
   while(writerIterator != WriterSet.end()) {
      if((*writerIterator)->write(data,length) == false) {
         success = false;
      }
      writerIterator++;
   }

   unsynchronized();
   return(true);
}


}
