// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Network Monitor Implementation                                   ####
// ####                                                                  ####
// #### Version 1.00  --  February 16, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "networkmonitor.h"
#include "streammonitorreport.h"
#include "thread.h"
#include "tools.h"

#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>

#include "in6.h"

extern "C" {
#include <pcap.h>
}


// Debug mode: Print all UDP and TCP connections
// #define DEBUG
// Print unknown protocol's number
// #define PRINT_UNKNOWN


namespace Coral {


// ###### Constructor #######################################################
NetworkMonitor::NetworkMonitor(StreamMonitorReport* smr = NULL, char* device = NULL)
   : Thread()
{
   resetReport();
   
   PCapDescriptor    = NULL;
   PCapInterfaceType = (card32)-1;
   PCapHandler       = NULL;
   PCapDevice        = NULL;

   // Check user ID
   uid_t uid = geteuid();
   if(uid != 0) {
      cerr << "ERROR: NetworkMonitor::NetworkMonitor() - Only root is allowed to capture network devices!" << endl;
      return;
   }

   // Get default network device name, if no device name is given.
   char ebuf[PCAP_ERRBUF_SIZE];
   if(device == NULL) {
      device = pcap_lookupdev(ebuf);
      if(device == NULL) {
         cerr << "ERROR: NetworkMonitor::NetworkMonitor() - Unable to get default device name:" << endl
              << "       " << ebuf << "!" << endl;
         return;
      }
   }

   // Open device in promiscuous mode, save DefaultSnaplen bytes of each
   // packet, save error string into ebuf.
   PCapDescriptor = pcap_open_live(device,DefaultSnaplen,1,0,ebuf);
   if(PCapDescriptor == NULL) {
      cerr << "ERROR: NetworkMonitor::NetworkMonitor() - Unable to open network device " << device << ":" << endl
           << "       " << ebuf << "!" << endl;
      return;
   }

   // Select handler for the interface type of the opened device.
   PCapInterfaceType = pcap_datalink(PCapDescriptor);
   switch(PCapInterfaceType) {
      case DLT_NULL:
         PCapHandler = &loopbackInterfaceHandler;
       break;
      case DLT_EN10MB:
         PCapHandler = &ethernetInterfaceHandler;
       break;
      default:
         cerr << "ERROR: NetworkMonitor::NetworkMonitor() - Interface type "
              << PCapInterfaceType << " not implemented!" << endl;
         pcap_close(PCapDescriptor);
         PCapDescriptor = NULL;
       break;
   }

   // Set device name
   PCapDevice = device;
   SMReport   = smr;
}


// ###### Destructor ########################################################
NetworkMonitor::~NetworkMonitor()
{
   stop();
   if(PCapDescriptor != NULL) {
      pcap_close(PCapDescriptor);
      PCapDescriptor = NULL;
   }
}


// ###### Check, if NetworkMonitor is ready #################################
bool NetworkMonitor::ready() const
{
   return(PCapDescriptor != NULL);
}


// ###### Lock monitor to access data #######################################
void NetworkMonitor::lock()
{
   synchronized();
}


// ###### Unlock monitor ####################################################
void NetworkMonitor::unlock()
{
   unsynchronized();
}


// ###### Reset report ######################################################
void NetworkMonitor::resetReport()
{
   synchronized();
   Report.reset();
   TimeStamp = getMicroTime();
   unsynchronized();
}


// ###### Get report ########################################################
void NetworkMonitor::getReport(NetworkMonitorReport* report)
{
   synchronized();
   Report.Interval  = getMicroTime() - TimeStamp;
   Report.TimeStamp = TimeStamp;
   memcpy(report,&Report,sizeof(Report));
   unsynchronized();
}


// ###### Get report and reset ##############################################
void NetworkMonitor::moveReport(NetworkMonitorReport* report)
{
   synchronized();
   getReport(report);
   resetReport();
   unsynchronized();
}


// ###### Thread's run() implementation #####################################
void NetworkMonitor::run()
{
   if(PCapDescriptor == NULL) {
      cerr << "ERROR: NetworkMonitor::run() - Start without opened device!"
           << endl;
      exit(1);
   }

   // Handle packet headers on opened device
   resetReport();
   if(pcap_loop(PCapDescriptor,-1,PCapHandler,(u_char*)this) < 0) {
      cerr << "ERROR: NetworkMonitor::run() - pcap_loop failed!" << endl;
      exit(1);
   }
   exit(0);
}


// ###### Handler for TCP packets ###########################################
void NetworkMonitor::handleTCP(StreamSrcDest* stream,
                               const tcphdr*  tcp,
                               const cardinal tcpLength,
                               const cardinal rawLength)
{
   Report.BytesTCP     += tcpLength;
   Report.BytesTCPData += (tcpLength - (4 * tcp->doff));
   Report.PacketsTCP++;

   stream->Source.Port      = tcp->source;
   stream->Destination.Port = tcp->dest;

#ifdef DEBUG
   char str[64];
   snprintf((char*)&str,sizeof(str),,
            "$%02x/$%05x",stream->TrafficClass,stream->FlowLabel);
   cout << "TCP " << stream->Source << " -> " << stream->Destination << "  " << str << endl;  
#endif

   if(SMReport != NULL) {
      SMReport->update("TCP",stream->Source,stream->Destination,stream->FlowLabel,
                       stream->TrafficClass,rawLength,(tcpLength - (4 * tcp->doff)));
   }
}


// ###### Handler for UDP packets ###########################################
void NetworkMonitor::handleUDP(StreamSrcDest* stream,
                               const udphdr*  udp,
                               const cardinal udpLength,
                               const cardinal rawLength)
{
   Report.BytesUDP     += udpLength;
   Report.BytesUDPData += (udpLength - sizeof(udphdr));
   Report.PacketsUDP++;

   stream->Source.Port      = udp->source;
   stream->Destination.Port = udp->dest;

#ifdef DEBUG
   char str[64];
   snprintf((char*)&str,sizeof(str),
           "$%02x/$%05x",stream->TrafficClass,stream->FlowLabel);
   cout << "UDP " << stream->Source << " -> " << stream->Destination << "  " << str << endl;
#endif

   if(SMReport != NULL) {
      SMReport->update("UDP",stream->Source,stream->Destination,stream->FlowLabel,
                       stream->TrafficClass,rawLength,udpLength - sizeof(udphdr));
   }
}


// ###### Handler for ICMP4 packets ##########################################
void NetworkMonitor::handleICMP4(StreamSrcDest* stream,
                                 const icmphdr* icmp,
                                 const cardinal icmpLength,
                                 const cardinal rawLength)
{
   Report.BytesICMP4 += icmpLength;
   Report.PacketsICMP4++;
   if(SMReport != NULL) {
      SMReport->update("ICMP4",stream->Source,stream->Destination,stream->FlowLabel,
                       stream->TrafficClass,rawLength,icmpLength);
   }
}


// ###### Handler for ICMP6 packets ##########################################
void NetworkMonitor::handleICMP6(StreamSrcDest*   stream,
                                 const icmp6_hdr* icmp,
                                 const cardinal   icmpLength,
                                 const cardinal   rawLength)
{
   Report.BytesICMP6 += icmpLength;
   Report.PacketsICMP6++;
   if(SMReport != NULL) {
      SMReport->update("ICMP6",stream->Source,stream->Destination,stream->FlowLabel,
                       stream->TrafficClass,rawLength,icmpLength);
   }
}


// ###### Handler for IPv4 packets ##########################################
void NetworkMonitor::ip4PacketHandler(const iphdr* ip, const cardinal length)
{
   // ====== Update byte and packet counter =================================
   Report.BytesIPv4 += (card64)length;
   Report.PacketsIPv4++;

   // ====== Set from and to addresses ======================================
   StreamSrcDest stream;
   stream.Source.Port      = 0; // In IP we do not know about ports!
   stream.Destination.Port = 0;
   for(cardinal i = 0;i < 5;i++) {
      stream.Source.Host[i]      = 0;
      stream.Destination.Host[i] = 0;
   }
   stream.Source.Host[5]      = 0xffff;
   stream.Destination.Host[5] = 0xffff;
   memcpy((void*)&stream.Source.Host[6],(void*)&ip->saddr,4);
   memcpy((void*)&stream.Destination.Host[6],(void*)&ip->daddr,4);
   stream.FlowLabel    = 0;
   stream.TrafficClass = ip->tos;

   // ====== Update this packet's QoS group's byte and packet counter =======
   cardinal tosIndex = TrafficClassValues::getIndexForTrafficClass(stream.TrafficClass);
   Report.ClassBytes[tosIndex] += (card64)length;
   Report.ClassPackets[tosIndex]++;

   // ====== Handle protocols ===============================================
   card8 protocol = ip->protocol;
   switch(protocol) {
      case IPPROTO_TCP:
        {
         const tcphdr* tcp = (tcphdr*)((char*)ip + sizeof(iphdr));
         const cardinal tcplength = length - sizeof(iphdr);
         handleTCP(&stream,tcp,tcplength,length);
        }
       break;
      case IPPROTO_UDP:
        {
         const udphdr* udp = (udphdr*)((char*)ip + sizeof(iphdr));
         const cardinal udplength = length - sizeof(iphdr);
         handleUDP(&stream,udp,udplength,length);
        }
       break;
      case IPPROTO_ICMP:
        const icmphdr* icmp = (icmphdr*)((char*)ip + sizeof(iphdr));
        const cardinal icmplength = length - sizeof(iphdr);
        handleICMP4(&stream,icmp,icmplength,length);
       break;
   }
}


// ###### Handler for IPv6 packets ##########################################
void NetworkMonitor::ip6PacketHandler(const ip6_hdr* ip, const cardinal length)
{
   // ====== Update byte and packet counter =================================
   Report.BytesIPv6 += (card64)length;
   Report.PacketsIPv6++;

   // ====== Set from and to addresses ======================================
   StreamSrcDest stream;
   stream.Source.Port      = 0; // In IP we do not know about ports!
   stream.Destination.Port = 0;
   memcpy((void*)&stream.Source.Host,&ip->ip6_src,16);
   memcpy((void*)&stream.Destination.Host,&ip->ip6_dst,16);
   stream.FlowLabel    = ntohl(ip->ip6_flow) & IPV6_FLOWINFO_FLOWLABEL;
   stream.TrafficClass = (ntohl(ip->ip6_flow) & IPV6_FLOWINFO_PRIORITY) >> 20;

   // ====== Update this packet's QoS group's byte and packet counter =======
   cardinal tosIndex = TrafficClassValues::getIndexForTrafficClass(stream.TrafficClass);
   Report.ClassBytes[tosIndex] += (card64)length;
   Report.ClassPackets[tosIndex]++;

   // ====== Handle protocols ===============================================
   card8 protocol = ip->ip6_nxt;
   switch(protocol) {
      case IPPROTO_TCP:
        {
         const tcphdr* tcp = (tcphdr*)((char*)ip + sizeof(ip6_hdr));
         const cardinal tcplength = length - sizeof(ip6_hdr);
         handleTCP(&stream,tcp,tcplength,length);
        }
       break;
      case IPPROTO_UDP:
        {
         const udphdr* udp = (udphdr*)((char*)ip + sizeof(ip6_hdr));
         const cardinal udplength = length - sizeof(ip6_hdr);
         handleUDP(&stream,udp,udplength,length);
        }
       break;
      case IPPROTO_ICMPV6:
        const icmp6_hdr* icmp = (icmp6_hdr*)((char*)ip + sizeof(ip6_hdr));
        const cardinal icmplength = length - sizeof(ip6_hdr);
        handleICMP6(&stream,icmp,icmplength,length);
       break;
   }
}


// ###### Handler for loopback frames #######################################
#define NULL_HDRLEN 4
void NetworkMonitor::loopbackInterfaceHandler(u_char*            userData,
                                              const pcap_pkthdr* pcapheader,
                                              const u_char*      packet)
{
   NetworkMonitor* monitor = (NetworkMonitor*)userData;
   const char*  payload    = (char*)packet + 4;
   const cardinal length     = pcapheader->len - NULL_HDRLEN;
   cardinal protocol;
   memcpy((char*)&protocol,packet,sizeof(protocol));
   protocol = ntohl(protocol) & 0xffff;

   monitor->synchronized();
   monitor->Report.BytesRaw += (card64)pcapheader->len;
   monitor->Report.PacketsRaw++;

   switch(protocol) {
      case ETH_P_IP:
         monitor->ip4PacketHandler((const iphdr*)payload,length);
        break;
      case ETH_P_IPV6:
         monitor->ip6PacketHandler((const ip6_hdr*)payload,length);
        break;
      default:
#ifdef PRINT_UNKNOWN
         printf("Unknown protocol %04x\n",protocol);
#endif
        break;
   }

   monitor->unsynchronized();
}
                        

// ###### Handler for ethernet frames #######################################
void NetworkMonitor::ethernetInterfaceHandler(u_char*            userData,
                                              const pcap_pkthdr* pcapheader,
                                              const u_char*      packet)
{
   NetworkMonitor*     monitor  = (NetworkMonitor*)userData;
   const char*         payload  = (char*)packet + sizeof(struct ether_header);
   const cardinal      length   = pcapheader->len - sizeof(struct ether_header);
   const ether_header* ethernet = (struct ether_header*)packet;
   const card16        protocol = ntohs(ethernet->ether_type);

   monitor->synchronized();
   monitor->Report.BytesRaw += (card64)pcapheader->len;
   monitor->Report.PacketsRaw++;

   switch(protocol) {
      case ETH_P_IP:
         monitor->ip4PacketHandler((const iphdr*)payload,length);
        break;
      case ETH_P_IPV6:
         monitor->ip6PacketHandler((const ip6_hdr*)payload,length);
        break;
      case ETH_P_ARP:
      case ETH_P_RARP:
       break;
      default:
#ifdef PRINT_UNKNOWN
         printf("Unknown protocol %04x\n",protocol);
#endif
        break;
   }

   monitor->unsynchronized();
}


}
