package bonn.infoIV.v3d2.rtp;


import bonn.infoIV.v3d2.util.ByteArrayToIntArray;
import bonn.infoIV.v3d2.util.IntArrayToByteArray;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.EOFException;

public class RTCPPacket
{
  public static final int MIN_PACKAGE_SIZE = 32;

  public RTCPHeader    MyHeader_         = null;
  public int[]         ApplicationData_  = null; // for PAYLOAD_TYPE_APP
  public RTCPReportBlock[] ReportBlocks_ = null; // for PAYLOAD_TYPE_RR



  public RTCPPacket(boolean padding, int rc, int ssrc, boolean noUse)
  {
    super();
    MyHeader_ = new RTCPHeader(padding, rc, ssrc, noUse);
  }


  public RTCPPacket(boolean padding, int rc, int ssrc, RTCPReportBlock report)
  {
    super();
    MyHeader_        = new RTCPHeader(padding, rc, ssrc, Control.NO_USE);
    ReportBlocks_    = new RTCPReportBlock[1];
    ReportBlocks_[0] = report;
  }


  public RTCPPacket(boolean padding, int rc, int ssrc, RTCPReportBlock[] reports)
  {
    super();
    MyHeader_      = new RTCPHeader(padding, rc, ssrc);
    ReportBlocks_  = reports;
  }


  public RTCPPacket(boolean padding, int src, int name)
  {
    super();
    MyHeader_ = new RTCPHeader(padding, src, name);
  }


  public RTCPPacket(boolean padding, int st, int src, int name, int data)
  {
    super();
    MyHeader_           = new RTCPHeader(padding, st, src, name);
    ApplicationData_    = new int[1];
    ApplicationData_[0] = data;
  }


  public RTCPPacket(boolean padding, int st, int src, int name, int[] data)
  {
    super();
    MyHeader_        = new RTCPHeader(padding, st, src, name);
    ApplicationData_ = data;
  }


  public RTCPPacket()
  {
    super();
    MyHeader_ = new RTCPHeader();
  }



  public void readPacket(DataInputStream in)
  {
    try
    {
      MyHeader_.readHeader(in);
      switch(MyHeader_.getPayloadType())
      {
        case RTCPHeader.PAYLOAD_TYPE_RR:
          ReportBlocks_ = new RTCPReportBlock[MyHeader_.getReceiverCount()];
          for(int i=0 ; i<ReportBlocks_.length ; i++)
	  {
            ReportBlocks_[i] = new RTCPReportBlock();
            ReportBlocks_[i].readReport(in);
	  }
          break;
        case RTCPHeader.PAYLOAD_TYPE_APP:
          ByteArrayToIntArray batia = new ByteArrayToIntArray(in);
          ApplicationData_ = new int[(int)(MyHeader_.getLength()-2)];
          batia.readInt(ApplicationData_);
          break;
        default:
          System.out.println("We do not support this RTCP PayloadType yet while writing: " +MyHeader_.getPayloadType());
      }
    }
    catch(Exception e)
    {
      if(e instanceof EOFException)
      {
        System.err.println("[ERROR] EOFExeption while reading RTPPacket: " +e.getMessage());
      }
      else System.err.println("[ERROR] reading RTCPPacket: " +e.getMessage());
    }
  }


  public void writePacket(DataOutputStream out)
  {
    try
    {
      MyHeader_.writeHeader(out);
      switch(MyHeader_.getPayloadType())
      {
        case RTCPHeader.PAYLOAD_TYPE_RR:
          for(int i=0 ; i<ReportBlocks_.length ; i++)
            ReportBlocks_[i].writeReport(out);
          break;
        case RTCPHeader.PAYLOAD_TYPE_APP:
          IntArrayToByteArray iatba = new IntArrayToByteArray(out);
          iatba.writeInt(ApplicationData_);
          break;
        default:
          System.out.println("We do not support this RTCP PayloadType yet while writing: " +MyHeader_.getPayloadType());
      }
      out.flush();
    }
    catch(IOException e) { System.err.println("[ERROR] writing RTPPacket: " +e.getMessage()); }
  }


  public void setHeader(RTCPHeader h) { MyHeader_ = h; }
  public RTCPHeader getHeader() { return MyHeader_; }

  public void setApplicationData(int[] data) { ApplicationData_ = data; }
  public int[] getApplicationData() { return ApplicationData_; }

  public void setReportBlocks(RTCPReportBlock[] reports) { ReportBlocks_ = reports; }
  public RTCPReportBlock[] getReportBlocks() { return ReportBlocks_; }

  public void setReportBlock(RTCPReportBlock report)
  { 
    ReportBlocks_ = null; 
    ReportBlocks_ = new RTCPReportBlock[1];
    ReportBlocks_[0] = report;
  }

  public void setReportBlockAt(RTCPReportBlock report, int index)
  {
    if (index>=ReportBlocks_.length)
    {
      System.err.println("[ERROR]: This Index is too big: " +index +" in RTPCPacket");
      return;
    }
    ReportBlocks_[index] = report; 
  }

  public RTCPReportBlock getReportBlock(int index)
  {
    if (index>=ReportBlocks_.length) return null;
    else return ReportBlocks_[index]; 
  }

  public byte[] getAsByteArray()
  {
    int[] a = null;
    int[] h = MyHeader_.getFullHeader();
    switch(MyHeader_.getPayloadType())
    {
      case RTCPHeader.PAYLOAD_TYPE_RR:
        a = new int[h.length+ReportBlocks_.length*RTCPReportBlock.REPORT_BLOCK_LENGTH];
        for(int i=0 ; i<h.length ; i++) a[i] = h[i];
        for(int i=0 ; i<ReportBlocks_.length ; i++)
        {
          for(int j=(h.length+ i   *RTCPReportBlock.REPORT_BLOCK_LENGTH) ; 
                  j<(h.length+(i+1)*RTCPReportBlock.REPORT_BLOCK_LENGTH) ; j++)
            a[j] = ReportBlocks_[i].getReportBlock()[j - (h.length + i*RTCPReportBlock.REPORT_BLOCK_LENGTH)];
        }
        return IntArrayToByteArray.toByte(a);
      case RTCPHeader.PAYLOAD_TYPE_APP:
        a = new int[h.length+ApplicationData_.length];
        for(int i=0 ; i<h.length ; i++) a[i] = h[i];
        for(int i=h.length ; i<(h.length+ApplicationData_.length) ; i++) {
           a[i] = ApplicationData_[i-h.length];
        }
        return(IntArrayToByteArray.toByte(a));
      default:
        return IntArrayToByteArray.toByte(h);

    }
  }
}

