// RTCP_Packet.hh --- RTCP_Packet Class Header File

#ifndef RTCP_PACKET_H
#define RTCP_PACKET_H

#include <stl_config.h>             // bool
#include <RTP/common.hh>            // Typedefs
#include <byte_order/byte_order.hh> // reverse()

enum RTCP_Type
{
  RTCP_SR   = 200,
  RTCP_RR   = 201,
  RTCP_SDES = 202,
  RTCP_BYE  = 203,
  RTCP_APP  = 204
};

enum RTCP_SDES_Type
{
  RTCP_SDES_END   = 0,
  RTCP_SDES_CNAME = 1,
  RTCP_SDES_NAME  = 2,
  RTCP_SDES_EMAIL = 3,
  RTCP_SDES_PHONE = 4,
  RTCP_SDES_LOC   = 5,
  RTCP_SDES_TOOL  = 6,
  RTCP_SDES_NOTE  = 7,
  RTCP_SDES_PRIV  = 8
};

class RTCP_Common_Header
{
private:
#ifndef BYTE_ORDER
#error "BYTE_ORDER undefined !"
#elif BYTE_ORDER == BIG_ENDIAN
  u_int16 v:2;		                  // Protocol Version
  u_int16 p:1;		                  // Padding Flag
  u_int16 c:5;		                  // Count varies by Packet Type (RC/SC)
#elif BYTE_ORDER == LITTLE_ENDIAN
  u_int16 c:5;		                  // Count varies by Packet Type (RC/SC)
  u_int16 p:1;		                  // Padding Flag
  u_int16 v:2;		                  // Protocol Version
#else
#error "BYTE_ORDER unknown !"
#endif
  u_int16 pt:8;		                  // RTCP Packet Type
  u_int16 len;                            // Packet length in words minus one
public:
  RTCP_Common_Header();
  u_int32 getVersion();
  u_int32 getPadding();
  u_int32 getCount();
  u_int32 getPacketType();
  u_int16 getLength();
  void setVersion(u_int32);
  void setPadding(u_int32);
  void setCount(u_int32);
  void setPacketType(u_int32);
  void setLength(u_int16);
};

class RTCP_Sender_Info_Block
{
private:
  u_int32 ntp_sec;                        // NTP Timestamp (most  significant word)
  u_int32 ntp_frac;                       // NTP Timestamp (least significant word)
  u_int32 rtp_ts;                         // RTP Timestamp
  u_int32 p_sent;                         // Packets sent
  u_int32 o_sent;                         // Octets sent
public:
  RTCP_Sender_Info_Block();
  u_int32 getNTPSeconds();
  u_int32 getNTPFraction();
  u_int32 getRTPTimeStamp();
  u_int32 getPacketsSent();
  u_int32 getOctetsSent();
  void setNTPSeconds(u_int32);
  void setNTPFraction(u_int32);
  void setRTPTimeStamp(u_int32);
  void setPacketsSent(u_int32);
  void setOctetsSent(u_int32);
};

class RTCP_Reception_Report_Block
{
private:
  u_int32 ssrc;                           // Data Source being reported
  u_int32 fraction:8;                     // Fraction lost since last SR/RR
  u_int32 lost:24;                        // Cumulative no. of Packets lost (signed!)
  u_int32 last_seq;                       // Extended last sequence no. received
  u_int32 jitter;                         // Interarrival Jitter
  u_int32 lsr;                            // Last SR Packet from this source
  u_int32 dlsr;                           // Delay since last SR Packet
public:
  RTCP_Reception_Report_Block();
  RTCP_Reception_Report_Block(u_int32);
  u_int32 getSSRC();
  float getFractionLost();
  u_int32 getPacketsLost();
  u_int32 getLastSeqNum();
  u_int32 getJitter();
  u_int32 getLSR();
  u_int32 getDLSR();
  void setSSRC(u_int32);
  void setFractionLost(u_int32);
  void setPacketsLost(u_int32);
  void setLastSeqNum(u_int32);
  void setJitter(u_int32);
  void setLSR(u_int32);
  void setDLSR(u_int32);
};

class RTCP_Source_Description_Item
{
private:
  u_int08 type;                           // Type of Item (RTCP_SDES_Type)
  u_int08 len;                            // Length of Item (in octets)
  char    data[1];                        // Text, not null-terminated
};

class RTCP_Source_Description_Chunk
{
private:
  u_int32                      src;       // SSRC/CSRC Identifier of source
  RTCP_Source_Description_Item item[1];   // Variable length SDES Item list
};

class RTCP_Packet : public RTCP_Common_Header
{

};

class RTCP_Report : public RTCP_Packet
{
private:
  u_int32 ssrc;                           // Sender or Receiver generating this report
public:
  RTCP_Report();
  u_int32 getSSRC();
  void setSSRC(u_int32);
};

class RTCP_Sender_Report : public RTCP_Report
{
public:
  RTCP_Sender_Info_Block si;              // Sender Info Block
  RTCP_Reception_Report_Block rr[1];      // Variable length RR list
public:
  RTCP_Sender_Report();
};

class RTCP_Receiver_Report : public RTCP_Report
{
public:
  RTCP_Reception_Report_Block rr[1];      // Variable length RR list
public:
  RTCP_Receiver_Report();
};

class RTCP_Source_Description : public RTCP_Packet
{
private:
  RTCP_Source_Description_Chunk chunk[1]; // Variable length SDES Chunk list
public:
  RTCP_Source_Description();
};

class RTCP_Bye : public RTCP_Packet
{
private:
  u_int32 src[1];                         // Variable length SSRC/CSRC list
public:
  RTCP_Bye();
};

class RTCP_App : public RTCP_Packet
{
public:
  u_int32 ssrc;
  char name[4];
  u_int32 val[4];
public:
  RTCP_App();
};

class RTCP_Union_Packet : public RTCP_Packet
{
protected:
  union
  {
    char a[sizeof(RTCP_Sender_Report)];
    char b[sizeof(RTCP_Receiver_Report)];
    char c[sizeof(RTCP_Source_Description)];
    char d[sizeof(RTCP_Bye)];
    char e[sizeof(RTCP_App)];
  };
public:
  RTCP_Union_Packet();
};

#endif
