// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### QSpectrumAnalyzer implementation                                 ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "wavaudioreader.h"
#include "audiodevice.h"
#include "audiodebug.h"
#include "audioconverter.h"
#include "timedthread.h"


#include <qapp.h>
#if QT_VERSION < 200
#error QT Version 2.00 or better required!!!
#endif
#include <qlayout.h>
#include <qpushbutton.h>
#include <qbuttongroup.h>
#include <qradiobutton.h>
#include <qtimer.h>
#include <qpainter.h>
#include <qgroupbox.h>
#include <qcheckbox.h>
#include <qplatinumstyle.h>
#include <qmainwindow.h>

#include "qspectrumanalyzer.h"


// ====== Include Qt's meta objects ======
#include "qspectrumanalyzer.moc"


namespace Coral {


// ###### Constructor #######################################################
QSpectrumAnalyzer::QSpectrumAnalyzer(SpectrumAnalyzer* analyzer,
                                     QWidget*          parent = NULL,
                                     const char*       name   = NULL)
   : QMainWindow(parent,name)
{
   Analyzer = analyzer;
   Max      = 1;

   // ====== Central widget =================================================
   QWidget* centralWidget = new QWidget(this);
   CHECK_PTR(centralWidget);
   QGridLayout* layout    = new QGridLayout(centralWidget,1,2);
   CHECK_PTR(layout);
   layout->setColStretch(0,0);
   layout->setColStretch(1,10);

   // ====== Control group ==================================================
   QGroupBox* controlGroup = new QGroupBox("Control",centralWidget);
   CHECK_PTR(controlGroup);
   layout->addWidget(controlGroup,0,0);
   QVBoxLayout* controlLayout = new QVBoxLayout(controlGroup,20,20);  
   CHECK_PTR(controlLayout);

   Pause = new QPushButton("Pause",controlGroup);
   CHECK_PTR(Pause);
   Pause->setToggleButton(TRUE);
   Pause->setOn(FALSE);
   controlLayout->addWidget(Pause);
   QObject::connect(Pause,SIGNAL(toggled(bool)),this,SLOT(pause(bool)));

   QPushButton* buttonClose = new QPushButton("Close",controlGroup);
   CHECK_PTR(buttonClose);
   controlLayout->addWidget(buttonClose);
   QObject::connect(buttonClose,SIGNAL(clicked()),this,SLOT(closeWindow()));

   Average = new QCheckBox("Average",controlGroup);
   CHECK_PTR(Average);
   Average->setChecked(TRUE);
   controlLayout->addWidget(Average);

   // ====== Timing group ===================================================
   QButtonGroup* radioGroup = new QButtonGroup(controlGroup);
   CHECK_PTR(radioGroup);
   controlLayout->addWidget(radioGroup);
   QVBoxLayout* radioLayout = new QVBoxLayout(radioGroup,5);
   CHECK_PTR(radioLayout);

   char str[16];
   for(cardinal i = 0;i < sizeof(QSpectrumAnalyzerTimings) / sizeof(card16);i++) {
      snprintf((char*)&str,sizeof(str),
              "%1.2f s",(double)QSpectrumAnalyzerTimings[i] / 1000.0);
      QRadioButton* radio = new QRadioButton(str,radioGroup);
      CHECK_PTR(radio);
      if(i == 1) {
         radio->setChecked(true);
      }
      radioLayout->addWidget(radio);
   }
   QObject::connect(radioGroup,SIGNAL(clicked(int)),this,SLOT(newInterval(int)));

   // ====== Analyzer display ===============================================
   QGroupBox* fourierGroup = new QGroupBox("Fast Fourier Spectrum Analyzer",centralWidget);
   CHECK_PTR(fourierGroup);
   layout->addWidget(fourierGroup,0,1);
   QGridLayout* fourierLayout = new QGridLayout(fourierGroup,2,1,20,20);
   CHECK_PTR(fourierLayout);
   PaintWidget1 = new QWidget(fourierGroup);
   CHECK_PTR(PaintWidget1);
   PaintWidget1->setMinimumWidth(300);
   PaintWidget1->setMinimumHeight(150);
   fourierLayout->addWidget(PaintWidget1,0,0);
   PaintWidget2 = new QWidget(fourierGroup);
   CHECK_PTR(PaintWidget2);
   PaintWidget2->setMinimumWidth(300);
   PaintWidget2->setMinimumHeight(150);
   fourierLayout->addWidget(PaintWidget2,1,0);

   setCentralWidget(centralWidget);
   setCaption("Spectrum Analyzer");

   Timer = new QTimer(this);
   CHECK_PTR(Timer);
   QObject::connect(Timer,SIGNAL(timeout()),this,SLOT(timerEvent()));
   Timer->start(QSpectrumAnalyzerTimings[1]);
   Timing = QSpectrumAnalyzerTimings[1];

   reset();
}


// ###### Destructor ########################################################
QSpectrumAnalyzer::~QSpectrumAnalyzer()
{
   delete Timer;
   Timer = NULL;
}


// ###### Draw spectrum bar #################################################
void QSpectrumAnalyzer::drawBar(QPainter*      painter,
                                const cardinal x,
                                const cardinal y,
                                const cardinal width,
                                const cardinal height,
                                const cardinal barValue)
{
   cardinal value = barValue;
   if(barValue > height)
      value = height;

   // ====== Remove area above bar =========================================
   painter->fillRect(x,y,width,height - value,PaintWidget1->backgroundColor());

   const cardinal y1 = y + height - value;
   const cardinal h1 = value / BarColors;

   // ====== Draw the bar ===================================================
   cardinal r = 255;
   cardinal g = 0;
   cardinal i;
   for(i = 0;i < BarColors - 1;i++) {
      r -= (255 / BarColors);
      g += (255 / BarColors);
      painter->fillRect(x,y1 + i * h1,width,h1,QColor(r,g,0));
   }   
   painter->fillRect(x,y1 + i * h1,width,y + height - (y1 + i * h1),QColor(0,255,0));
}


// ###### Show spectrum #####################################################
void QSpectrumAnalyzer::showSpectrum(QWidget*        paintWidget,
                                     const cardinal* array)
{
   QPainter painter;
   const cardinal x      = 0;
   const cardinal y      = 0;
   const cardinal width  = paintWidget->width();
   const cardinal height = paintWidget->height() - 3;

   painter.begin(paintWidget);
   painter.fillRect(x,y+height,width,3,QColor(10,10,10));

   // ====== Calculate maximum value ========================================
   cardinal values[Bars];
   Max = (cardinal)(double)(Max * 0.98);
   if(Max < 1)
      Max = 1;
   for(cardinal i = 0;i < Bars;i++) {
      values[i] = array[i];
      if(values[i] > Max)
         Max = values[i];
   }

   // ====== Draw Bars ======================================================
   cardinal step = width / Bars;
   cardinal sx   = x;
   for(cardinal i = 0;i < Bars;i++) {
      const cardinal value = (values[i] * height) / Max;
      drawBar(&painter,sx,y,step,height,value);
      sx += step;
   }

   // ====== Draw average line ==============================================
   if(Average->isOn()) {
      step = width / Bars;
      sx = x;
      painter.setPen(QPen(QColor(80,80,255),3));
      for(cardinal i = 0;i < Bars;i += AverageSteps) {
         cardinal sum = values[i];
         cardinal j;
         for(j = 1;j < AverageSteps;j++) {
            if((i + j >= Bars))           
               break;
            sum += values[i + j];
         }
         sum /= AverageSteps;

         if(i == 0) {
            painter.moveTo(sx,y + height - sum);
         } else {      
            painter.lineTo(sx,y + height - sum);
         }
         sx += AverageSteps * step;
         if(sx > Bars * step) {
            sx = Bars * step;
         }
         painter.lineTo(sx,y + height - sum);
      }
   }

   painter.end();
}


// ###### Timer event slot ##################################################
void QSpectrumAnalyzer::timerEvent()
{
   if(Analyzer->getSpectrum((cardinal*)&ArrayL,(cardinal*)&ArrayR,Bars)) {
      showSpectrum(PaintWidget1,ArrayL);
      showSpectrum(PaintWidget2,ArrayR);
   }
}


// ###### Paint event slot ##################################################
void QSpectrumAnalyzer::paintEvent(QPaintEvent*)
{
   showSpectrum(PaintWidget1,ArrayL);
   showSpectrum(PaintWidget2,ArrayR);
}


// ###### Change update interval ############################################
void QSpectrumAnalyzer::newInterval(int index)
{
   if((cardinal)index < sizeof(QSpectrumAnalyzerTimings) / sizeof(card16)) {
      Timing = QSpectrumAnalyzerTimings[index];
      if(Pause->isOn()) {
         Pause->setOn(FALSE);
      }
      Timer->changeInterval(QSpectrumAnalyzerTimings[index]);
   }
}


// ###### Pause displaying the spectrum #####################################
void QSpectrumAnalyzer::pause(bool on)
{
   if(on) {
      Timer->stop();
   }
   else {
      Timer->start(Timing);
   }
}


// ###### Reset spectrum analyzer ###########################################
void QSpectrumAnalyzer::reset()
{
   Analyzer->sync();
   for(cardinal i = 0;i < Bars;i++) {
      ArrayL[i] = 0;
      ArrayR[i] = 0;
   }
   repaint();
}


// ###### Close window ######################################################
void QSpectrumAnalyzer::closeWindow()
{
   emit closeSpectrumAnalyzer();
}


// ###### Close window ######################################################
void QSpectrumAnalyzer::closeEvent(QCloseEvent* event)
{
   emit closeSpectrumAnalyzer();
}


}
